/*
 * Copyright (c) 2005-2006 Jason Katz-Brown and John O'Laughlin.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The names of the authors may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef QUACKLE_ENDGAME_H
#define QUACKLE_ENDGAME_H

#include <fstream>
#include <math.h>
#include <vector>

#include "alphabetparameters.h"
#include "game.h"

namespace Quackle
{

struct EndgameMove
{
	EndgameMove(const Move &_move) : move(_move) { }
	Move move;
	double optimistic;
	double pessimistic;
	double estimated;
	bool outplay;
};

class EndgameMoveList : public vector<EndgameMove>
{
public:
	static bool optimisticComparator(const EndgameMove &move1, const EndgameMove &move2);
};

class Endgame
{
public:
	// constructs a new endgame solver
	Endgame();
	~Endgame();

	// Find the best move in this position. Also initializes the
	// move list, rack, resets numbers, and closes logfile.
	void setPosition(const GamePosition &position);

	// get access to the position that starts each playahead of the
	// endgame; use to rechange rack or scores etcetera
	GamePosition &currentPosition();
	const GamePosition &currentPosition() const;

	// If logfile is an empty string, logging is disabled.
	// If logfile is the same logfile as currently set, nothing
	// happens. If it is different, old logfile is closed if it
	// was open. If append is false, this destroys file contents
	// already in logfile.
	void setLogfile(const string &logfile, bool append = true);
	string logfile() const;

	// append message to logfile if one is open
	void logMessage(const UVString &message);

	bool isLogging() const;
	void closeLogfile();

	// Set moves to include in solution.
	void setIncludedMoves(const MoveList &moves);

	// include only currently included moves that are within
	// equityThreshold points below the best play and cap at
	// maxNumberOfMoves
	void pruneTo(double equityThreshold, int maxNumberOfMoves);

	// return a list of moves, sorted by estimated equity
	MoveList moves() const;
	
	// return the move list
	const EndgameMoveList &endgameMoves() const;

	// Return the best move
	Move solve();
	double disappoint(EndgameMove hope, double bestPessimistic);

protected:
	void writeLogHeader();
	void writeLogFooter();

	UVOFStream m_logfileStream;
	string m_logfile;
	bool m_logfileIsOpen;
	bool m_hasHeader;
	UVString m_xmlIndent;

	Game m_originalGame;
	Game m_endgameGame;

	EndgameMoveList m_endgameMoves;
};

inline GamePosition &Endgame::currentPosition()
{
	return m_originalGame.currentPosition();
}

inline const GamePosition &Endgame::currentPosition() const
{
	return m_originalGame.currentPosition();
}

inline string Endgame::logfile() const
{
	return m_logfile;
}

inline bool Endgame::isLogging() const
{
	return m_logfileIsOpen;
}

inline const EndgameMoveList &Endgame::endgameMoves() const
{
	return m_endgameMoves;
}

}

UVOStream& operator<<(UVOStream& o, const Quackle::EndgameMove &move);
UVOStream& operator<<(UVOStream& o, const Quackle::EndgameMoveList &move);

#endif
