/*
 * Copyright (c) 2005-2006 Jason Katz-Brown and John O'Laughlin.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The names of the authors may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "board.h"
#include "datamanager.h"
#include "evaluator.h"
#include "strategyparameters.h"
#include "catchall.h"

using namespace Quackle;

double Evaluator::equity(const GamePosition &position, const Move &move) const
{
	return move.score;
}

double Evaluator::playerConsideration(const GamePosition &position, const Move &move) const
{
	return 0;
}

double Evaluator::sharedConsideration(const GamePosition &position, const Move &move) const
{
	return 0;
}

double Evaluator::leaveValue(const LetterString &leave) const
{
	return 0;
}

////////////

double ScorePlusLeaveEvaluator::equity(const GamePosition &position, const Move &move) const
{
	return playerConsideration(position, move) + sharedConsideration(position, move) + move.score;
}

double ScorePlusLeaveEvaluator::playerConsideration(const GamePosition &position, const Move &move) const
{
	LetterString leave;
	LetterString dummy = position.currentPlayer().rack().tiles();

	// TODO this chunk is superfluous; remove
	// (should just use rack - move no?)
	const LetterString::const_iterator end = move.tiles().end();
	int i = 0;
	for (LetterString::const_iterator it = move.tiles().begin(); it != end; ++it, ++i)
	{
		Letter tile = QUACKLE_NULL_MARK;

		if (move.action == Move::Exchange)
		{
			tile = *it;
		}
		else if (move.action == Move::Place)
		{
			int col, row;
			if (move.horizontal)
			{
				col = move.startcol + i;
				row = move.startrow;
			}
			else
			{
				col = move.startcol;
				row = move.startrow + i;
			}

			if (position.board().letter(row, col) == QUACKLE_NULL_MARK)
			{
				if (QUACKLE_ALPHABET_PARAMETERS->isPlainLetter(*it))
					tile = *it;
				else
					tile = QUACKLE_BLANK_MARK;
			}
		}

		if (tile != QUACKLE_NULL_MARK)
		{
			const LetterString::iterator dummyEnd(dummy.end());
			for (LetterString::iterator dummyIt = dummy.begin(); dummyIt != dummyEnd; ++dummyIt)
			{
				if (*dummyIt == tile)
				{
					*dummyIt = QUACKLE_NULL_MARK;
					break;
				}
			}
		}
	}

#ifdef DEBUG_BOARD
	UVcout << "move: " << move << ", rack: " << position.currentPlayer().rack() << ", dummy: " << QUACKLE_ALPHABET_PARAMETERS->userVisible(dummy) << endl;
#endif

	const LetterString::const_iterator dummyEnd(dummy.end());
	for (LetterString::const_iterator dummyIt = dummy.begin(); dummyIt != dummyEnd; ++dummyIt)
		if (*dummyIt != QUACKLE_NULL_MARK)
			leave += *dummyIt;

	return leaveValue(leave);
}

double ScorePlusLeaveEvaluator::sharedConsideration(const GamePosition &position, const Move &move) const
{
	return 0;
}

double ScorePlusLeaveEvaluator::leaveValue(const LetterString &leave) const
{
	if (!QUACKLE_STRATEGY_PARAMETERS->isInitialized())
		return 0;

	double value = 0;
	double synergy = 0;
	LetterString uniqleave;

	const LetterString::const_iterator leaveEnd(leave.end());

	if (!leave.empty())
	{
		for (LetterString::const_iterator leaveIt = leave.begin(); leaveIt != leaveEnd; ++leaveIt)
			value += QUACKLE_STRATEGY_PARAMETERS->tileWorth(*leaveIt);

		for (int i = 0; i < leave.length() - 1; ++i)
			if (leave[i] == leave[i + 1])
				value += QUACKLE_STRATEGY_PARAMETERS->syn2(leave[i], leave[i]);

		uniqleave += leave[0];
		for (int i = 1; i < leave.length(); ++i)
			if (uniqleave[uniqleave.length() - 1] != leave[i])
				uniqleave += leave[i];

		if (uniqleave.length() >= 2)
		{
			for (int i = 0; i < uniqleave.length() - 1; ++i)
				for (int j = i + 1; j < uniqleave.length(); ++j)
					synergy += QUACKLE_STRATEGY_PARAMETERS->syn2(uniqleave[i], uniqleave[j]);

			// TODO handle the Q
			/*
			bool holdingq = false;
			for (int i = 0; i < leave.length(); ++i) {
				if (leave[i] == 'Q') {
					holdingq = true;
				}
			}
			*/

			// if ((synergy > 3.0) && !holdingq) {
			//    synergy += 2.5 * (synergy - 3.0);
			// }

			value += synergy;
		}    
	}

	int vowels = 0;
	int cons = 0;

	for (LetterString::const_iterator leaveIt = leave.begin(); leaveIt != leaveEnd; ++leaveIt)
	{
		if (*leaveIt != QUACKLE_BLANK_MARK)
		{
			if (QUACKLE_ALPHABET_PARAMETERS->isVowel(*leaveIt))
				vowels++;
			else
				cons++;
		}
	} 

	const float vcvalues[8][8] =
	{
		{  0.0,   0.0,  -1.0,  -2.5,  -5.0,  -8.5, -13.5, -18.5},
		{ -1.0,   0.0,   0.5,   0.0,  -1.5,  -5.0, -10.0,   0.0},
		{ -3.5,  -1.0,   0.5,   1.5,  -1.5,  -3.0,   0.0,   0.0},
		{ -7.0,  -3.5,  -0.5,   2.5,   0.0,   0.0,   0.0,   0.0},
		{-10.0,  -6.5,  -3.0,   0.0,   0.0,   0.0,   0.0,   0.0},
		{-13.5, -11.5,  -8.0,   0.0,   0.0,   0.0,   0.0,   0.0},
		{-18.5, -16.5,   0.0,   0.0,   0.0,   0.0,   0.0,   0.0},
		{-23.5,   0.0,   0.0,   0.0,   0.0,   0.0,   0.0,   0.0},
	};

#ifdef DEBUG_BOARD
	UVcout << QUACKLE_ALPHABET_PARAMETERS->userVisible(leave) << " has " << vowels << " vowels, " << cons << " cons.  value of " << vcvalues[vowels][cons] << endl;
#endif

	value += vcvalues[vowels][cons];

	if (value < -40)
		value = -40;

#ifdef DEBUG_BOARD
	UVcout << "leave " << QUACKLE_ALPHABET_PARAMETERS->userVisible(leave) << " worth " << value << " uniq " << QUACKLE_ALPHABET_PARAMETERS->userVisible(uniqleave) << " synergy " << synergy << endl;
#endif

	return value;
}

