/*
 * Copyright (c) 2005-2006 Jason Katz-Brown and John O'Laughlin.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The names of the authors may not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef QUACKLE_SIM_H
#define QUACKLE_SIM_H

#include <fstream>
#include <math.h>
#include <vector>

#include "alphabetparameters.h"
#include "game.h"

namespace Quackle
{

struct AveragedValue
{
	// new zeroed value
	AveragedValue()
		: m_valueSum(0), m_squaredValueSum(0), m_incorporatedValues(0)
	{
	}

	void incorporateValue(double newValue);

	// zero everything
	void clear();

	long double valueSum() const;
	long double squaredValueSum() const;
	long int incorporatedValues() const;

	// valueSum() / incorporatedValues() or zero
	// if there have been no incorporated values
	double averagedValue() const;

	// sqrt((incorporatedValues() * squaredValueSum() - valueSum()^2) / (incorporatedValues() * (incorporatedValues() - 1)))
	double standardDeviation() const;

private:
	long double m_valueSum;
	long double m_squaredValueSum;
	long int m_incorporatedValues;
};

inline void AveragedValue::incorporateValue(double newValue)
{
	m_valueSum += newValue;
	m_squaredValueSum += newValue * newValue;
	++m_incorporatedValues;
}

inline long double AveragedValue::valueSum() const
{
	return m_valueSum;
}

inline long double AveragedValue::squaredValueSum() const
{
	return m_squaredValueSum;
}

inline double AveragedValue::averagedValue() const
{
	return m_incorporatedValues == 0? 0 : (m_valueSum / m_incorporatedValues);
}

inline long int AveragedValue::incorporatedValues() const
{
	return m_incorporatedValues;
}

typedef vector<AveragedValue> AveragedValueList;

struct Level
{
	// expand the scores list to be at least number long
	void setNumberScores(int number);

	AveragedValueList scores;
};

typedef vector<Level> LevelList;

struct SimmedMove
{
	SimmedMove(const Move &_move) : move(_move), m_includeInSimulation(true) { }

	double calculateEquity() const;

	// expand the levels list to be at least number long
	void setNumberLevels(int number);

	// clear all level values
	void clear();

	bool includeInSimulation() const;
	void setIncludeInSimulation(bool includeInSimulation);

	// average wins value * 100
	double winPercentage() const;

	Move move;
	LevelList levels;
	AveragedValue residual;
	AveragedValue gameSpread;
	AveragedValue wins;

private:
	bool m_includeInSimulation;
};

inline bool SimmedMove::includeInSimulation() const
{
	return m_includeInSimulation;
}

inline void SimmedMove::setIncludeInSimulation(bool includeInSimulation)
{
	m_includeInSimulation = includeInSimulation;
}

inline double SimmedMove::winPercentage() const
{
	return wins.averagedValue() * 100;
}

typedef vector<SimmedMove> SimmedMoveList;

class Simulator
{
public:
	// constructs a new simulator
	Simulator();
	~Simulator();

	// Simulate moves on this position. Also initializes the
	// move list, rack, resets numbers, and closes logfile.
	void setPosition(const GamePosition &position);

	// get access to the position that starts each playahead of the
	// simulation; use to rechange rack or scores etcetera
	GamePosition &currentPosition();
	const GamePosition &currentPosition() const;

	// If logfile is an empty string, logging is disabled.
	// If logfile is the same logfile as currently set, nothing
	// happens. If it is different, old logfile is closed if it
	// was open. If append is false, this destroys file contents
	// already in logfile.
	void setLogfile(const string &logfile, bool append = true);
	string logfile() const;

	// append message to logfile if one is open
	void logMessage(const UVString &message);

	bool isLogging() const;
	void closeLogfile();

	// Set moves to include in simulation. Moves that
	// are in the simmed list now that are not in this given
	// list still live in the simmed list but are not iterated thru
	// during simulation. Moves that were not in the simmed list are added
	// and level values zeroed.
	void setIncludedMoves(const MoveList &moves);

	// include only currently included moves that are within
	// equityThreshold points below the best play and cap at
	// maxNumberOfMoves
	void pruneTo(double equityThreshold, int maxNumberOfMoves);

	// if ignore is true, oppos will always pass in simulation
	void setIgnoreOppos(bool ignore);
	bool ignoreOppos() const;

	// set values for all levels of all moves to zero
	void resetNumbers();

	// Run a chunk of the simulation.
	// If plies is negative, simulation runs to end of game.
	// Iterations is how many iterations to run before returning;
	// more iterations can be computed and incorporated by recalling 
	// simulate().
	void simulate(int plies, int iterations);

	// simulate one iteration
	void simulate(int plies);

	// Set oppo's racks to something random.
	// Possibly inference-aided randomness.
	void randomizeOppoRacks();

	// set drawing order for the first refill
	void randomizeDrawingOrder();

	// returns maximal number of iterations over all moves since
	// resetting numbers
	int iterations() const;

	bool hasSimulationResults() const;

	// full simulation information
	const SimmedMoveList &simmedMoves() const;

	// Return the moves sorted by simulated equity.
	// If prune is true, does not include plays that aren't included
	// in simulation anymore.
	MoveList moves(bool prune = false) const;

	const SimmedMove &simmedMoveForMove(const Move &move) const;

protected:
	void writeLogHeader();
	void writeLogFooter();

	UVOFStream m_logfileStream;
	string m_logfile;
	bool m_logfileIsOpen;
	bool m_hasHeader;
	UVString m_xmlIndent;

	Game m_originalGame;
	Game m_simulatedGame;

	SimmedMoveList m_simmedMoves;

	int m_iterations;
	bool m_ignoreOppos;
};

inline GamePosition &Simulator::currentPosition()
{
	return m_originalGame.currentPosition();
}

inline const GamePosition &Simulator::currentPosition() const
{
	return m_originalGame.currentPosition();
}

inline string Simulator::logfile() const
{
	return m_logfile;
}

inline bool Simulator::isLogging() const
{
	return m_logfileIsOpen;
}

inline void Simulator::setIgnoreOppos(bool ignore)
{
	m_ignoreOppos = ignore;
}

inline bool Simulator::ignoreOppos() const
{
	return m_ignoreOppos;
}

inline int Simulator::iterations() const
{
	return m_iterations;
}

inline bool Simulator::hasSimulationResults() const
{
	return m_iterations > 0;
}

inline const SimmedMoveList &Simulator::simmedMoves() const
{
	return m_simmedMoves;
}

}

UVOStream& operator<<(UVOStream& o, const Quackle::AveragedValue &value);
UVOStream& operator<<(UVOStream& o, const Quackle::Level &level);
UVOStream& operator<<(UVOStream& o, const Quackle::SimmedMove &move);
UVOStream& operator<<(UVOStream& o, const Quackle::SimmedMoveList &move);

#endif
